from typing import Annotated

from pydantic import Field

from phylogenie.generators.alisim import AliSimDatasetGenerator
from phylogenie.generators.dataset import DatasetGenerator
from phylogenie.generators.trees import (
    BDEITreeDatasetGenerator,
    BDSSTreeDatasetGenerator,
    BDTreeDatasetGenerator,
    CanonicalTreeDatasetGenerator,
    EpidemiologicalTreeDatasetGenerator,
    FBDTreeDatasetGenerator,
    TreeDatasetGeneratorConfig,
)

DatasetGeneratorConfig = Annotated[
    TreeDatasetGeneratorConfig | AliSimDatasetGenerator,
    Field(discriminator="data_type"),
]

__all__ = [
    "DatasetGeneratorConfig",
    "DatasetGenerator",
    "AliSimDatasetGenerator",
    "CanonicalTreeDatasetGenerator",
    "EpidemiologicalTreeDatasetGenerator",
    "FBDTreeDatasetGenerator",
    "BDTreeDatasetGenerator",
    "BDEITreeDatasetGenerator",
    "BDSSTreeDatasetGenerator",
]



from collections.abc import Callable
from dataclasses import dataclass

from phylogenie.skyline import SkylineParameter

Event = Callable[[float], None]


@dataclass
class StochasticEvent:
    fn: Event
    reactants: dict[str, int]
    rate: SkylineParameter


@dataclass
class TimedEvent:
    fn: Event
    times: list[float]



from collections.abc import Sequence

from phylogenie.skyline import SkylineParameterLike, skyline_parameter
from phylogenie.tree_node import TreeNode
from phylogenie.treesimulator.events import StochasticEvent, TimedEvent
from phylogenie.treesimulator.model import Model

STATE_KEY = "state"


def _get_node_name(node_id: int, state: str) -> str:
    return f"{node_id}|{state}"


class UnboundedPopulationModel(Model):
    def __init__(
        self, init_state: str, max_time: float | None = None, seed: int | None = None
    ):
        super().__init__(max_time=max_time, seed=seed)
        self._init_state = init_state
        self.init()

    def init(self):
        super().init()
        self._next_node_id = 0
        self._sampled: set[str] = set()
        self._times: dict[TreeNode, float] = {}
        self._nodes: dict[int, TreeNode] = {}
        seed_individual = self._get_new_individual(self._init_state)
        self._tree = self._nodes[seed_individual]

    @property
    def tree_size(self) -> int:
        return len(self._sampled)

    def _get_new_node(self, state: str) -> TreeNode:
        self._next_node_id += 1
        node = TreeNode(_get_node_name(self._next_node_id, state))
        node[STATE_KEY] = state
        return node

    def _get_new_individual(self, state: str) -> int:
        new_id = self.create_new_individual(state)
        node = self._get_new_node(state)
        self._nodes[new_id] = node
        return new_id

    def _set_branch_length(self, node: TreeNode, time: float) -> None:
        if node.branch_length is not None:
            raise ValueError(f"Branch length of node {node.name} is already set.")
        parent_time = 0.0 if node.parent is None else self._times[node.parent]
        node.branch_length = time - parent_time
        self._times[node] = time

    def _stem(self, individual: int, time: float) -> None:
        node = self._nodes[individual]
        self._set_branch_length(node, time)
        stem_node = self._get_new_node(self._states[individual])
        node.add_child(stem_node)
        self._nodes[individual] = stem_node

    def remove(self, individual: int, time: float) -> None:
        self.remove_individual(individual)
        self._set_branch_length(self._nodes[individual], time)

    def migrate(self, individual: int, new_state: str, time: float) -> None:
        self.migrate_individual(individual, new_state)
        self._stem(individual, time)

    def birth_from(self, parent_individual: int, new_state: str, time: float) -> int:
        new_individual = self._get_new_individual(new_state)
        parent_node = self._nodes[parent_individual]
        new_node = self._nodes[new_individual]
        parent_node.add_child(new_node)
        self._stem(parent_individual, time)
        return new_individual

    def sample(self, individual: int, time: float, removal: bool) -> None:
        if removal:
            self._sampled.add(self._nodes[individual].name)
            self.remove(individual, time)
        else:
            sample_node = self._get_new_node(self._states[individual])
            sample_node.branch_length = 0.0
            self._sampled.add(sample_node.name)
            self._nodes[individual].add_child(sample_node)
            self._stem(individual, time)

    def get_tree(self) -> TreeNode | None:
        tree = self._tree.copy()
        for node in list(tree.iter_postorder()):
            if node.name not in self._sampled and not node.children:
                if node.parent is None:
                    return None
                else:
                    node.parent.remove_child(node)
            elif len(node.children) == 1:
                (child,) = node.children
                child.update_parent(node.parent)
                child.branch_length += node.branch_length  # pyright: ignore
                if node.parent is None:
                    return child
                else:
                    node.parent.remove_child(node)
        return tree

    # --------------
    # Event handlers
    # --------------

    def add_birth_event(self, state: str, child_state: str, rate: SkylineParameterLike):
        def birth_event(time: float):
            self.birth_from(self.draw_individual(state), child_state, time)

        self._stochastic_events.append(
            StochasticEvent(
                fn=birth_event, reactants={state: 1}, rate=skyline_parameter(rate)
            )
        )

    def add_death_event(self, state: str, rate: SkylineParameterLike):
        def death_event(time: float):
            self.remove(self.draw_individual(state), time)

        self._stochastic_events.append(
            StochasticEvent(
                fn=death_event, reactants={state: 1}, rate=skyline_parameter(rate)
            )
        )

    def add_migration_event(
        self, state: str, target_state: str, rate: SkylineParameterLike
    ):
        def migration_event(time: float):
            self.migrate(self.draw_individual(state), target_state, time)

        self._stochastic_events.append(
            StochasticEvent(
                fn=migration_event,
                reactants={state: 1},
                rate=skyline_parameter(rate),
            )
        )

    def add_sampling_event(
        self,
        state: str,
        removal: bool,
        rate: SkylineParameterLike,
    ):
        def sampling_event(time: float):
            self.sample(self.draw_individual(state), time, removal)

        self._stochastic_events.append(
            StochasticEvent(
                fn=sampling_event, reactants={state: 1}, rate=skyline_parameter(rate)
            )
        )

    def add_timed_sampling_event(
        self,
        state: str,
        proportion: float,
        removal: bool,
        times: Sequence[float],
    ):
        def timed_sampling_event(time: float):
            for individual in self.get_individuals(state):
                if self._rng.random() < proportion:
                    self.sample(individual, time, removal)

        self._timed_events.append(
            TimedEvent(fn=timed_sampling_event, times=sorted(times))
        )



from phylogenie.draw import (
    draw_colored_dated_tree_categorical,
    draw_colored_dated_tree_continuous,
    draw_colored_tree_categorical,
    draw_colored_tree_continuous,
    draw_dated_tree,
    draw_tree,
)
from phylogenie.generators import (
    AliSimDatasetGenerator,
    BDEITreeDatasetGenerator,
    BDSSTreeDatasetGenerator,
    BDTreeDatasetGenerator,
    CanonicalTreeDatasetGenerator,
    DatasetGenerator,
    DatasetGeneratorConfig,
    EpidemiologicalTreeDatasetGenerator,
    FBDTreeDatasetGenerator,
    TreeDatasetGeneratorConfig,
)
from phylogenie.io import dump_fasta, dump_newick, load_fasta, load_newick, load_nexus
from phylogenie.main import run
from phylogenie.msa import MSA, Sequence
from phylogenie.skyline import (
    SkylineMatrix,
    SkylineMatrixCoercible,
    SkylineParameter,
    SkylineParameterLike,
    SkylineVector,
    SkylineVectorCoercible,
    SkylineVectorLike,
    skyline_matrix,
    skyline_parameter,
    skyline_vector,
)
from phylogenie.tree_node import TreeNode
from phylogenie.treesimulator import (
    EXPOSED_STATE,
    INFECTIOUS_STATE,
    SUPERSPREADER_STATE,
    Model,
    UnboundedPopulationModel,
    generate_trees,
    get_BD_model,
    get_BDEI_model,
    get_BDSS_model,
    get_canonical_model,
    get_epidemiological_model,
    get_FBD_model,
    simulate_tree,
)

__all__ = [
    "draw_colored_dated_tree_categorical",
    "draw_colored_dated_tree_continuous",
    "draw_colored_tree_categorical",
    "draw_colored_tree_continuous",
    "draw_dated_tree",
    "draw_tree",
    "AliSimDatasetGenerator",
    "BDEITreeDatasetGenerator",
    "BDSSTreeDatasetGenerator",
    "BDTreeDatasetGenerator",
    "CanonicalTreeDatasetGenerator",
    "DatasetGenerator",
    "DatasetGeneratorConfig",
    "EpidemiologicalTreeDatasetGenerator",
    "FBDTreeDatasetGenerator",
    "TreeDatasetGeneratorConfig",
    "dump_fasta",
    "dump_newick",
    "load_fasta",
    "load_newick",
    "load_nexus",
    "run",
    "MSA",
    "Sequence",
    "SkylineMatrix",
    "SkylineMatrixCoercible",
    "SkylineParameter",
    "SkylineParameterLike",
    "SkylineVector",
    "SkylineVectorCoercible",
    "SkylineVectorLike",
    "skyline_matrix",
    "skyline_parameter",
    "skyline_vector",
    "TreeNode",
    "EXPOSED_STATE",
    "INFECTIOUS_STATE",
    "SUPERSPREADER_STATE",
    "Model",
    "UnboundedPopulationModel",
    "generate_trees",
    "get_BD_model",
    "get_BDEI_model",
    "get_BDSS_model",
    "get_canonical_model",
    "get_epidemiological_model",
    "get_FBD_model",
    "simulate_tree",
]



from phylogenie.treesimulator.gillespie import generate_trees, simulate_tree
from phylogenie.treesimulator.model import Model
from phylogenie.treesimulator.unbounded_population import (
    EXPOSED_STATE,
    INFECTIOUS_STATE,
    SUPERSPREADER_STATE,
    UnboundedPopulationModel,
    get_BD_model,
    get_BDEI_model,
    get_BDSS_model,
    get_canonical_model,
    get_epidemiological_model,
    get_FBD_model,
)

__all__ = [
    "EXPOSED_STATE",
    "INFECTIOUS_STATE",
    "SUPERSPREADER_STATE",
    "get_BD_model",
    "get_BDEI_model",
    "get_BDSS_model",
    "get_canonical_model",
    "get_epidemiological_model",
    "get_FBD_model",
    "generate_trees",
    "simulate_tree",
    "Model",
    "UnboundedPopulationModel",
]



import re
from typing import Any

import numpy as np
from numpy.random import Generator

import phylogenie._typeguards as tg
import phylogenie._typings as pgt
import phylogenie.generators._configs as cfg
import phylogenie.generators._typeguards as ctg
from phylogenie.skyline import (
    SkylineMatrix,
    SkylineMatrixCoercible,
    SkylineParameter,
    SkylineParameterLike,
    SkylineVector,
    SkylineVectorCoercible,
)
from phylogenie.treesimulator import UnboundedPopulationModel


def eval_expression(
    expression: str, data: dict[str, Any], extra_globals: dict[str, Any] | None = None
) -> Any:
    if extra_globals is None:
        extra_globals = {}
    return np.array(
        eval(
            expression,
            {"np": np, **{k: np.array(v) for k, v in data.items()}, **extra_globals},
        )
    ).tolist()


def integer(x: cfg.Integer, data: dict[str, Any]) -> int:
    if isinstance(x, str):
        e = eval_expression(x, data)
        if isinstance(e, int):
            return e
        raise ValueError(
            f"Expression '{x}' evaluated to {e} of type {type(e)}, expected an int."
        )
    return x


def scalar(x: cfg.Scalar, data: dict[str, Any]) -> pgt.Scalar:
    if isinstance(x, str):
        e = eval_expression(x, data)
        if isinstance(e, pgt.Scalar):
            return e
        raise ValueError(
            f"Expression '{x}' evaluated to {e} of type {type(e)}, expected a scalar."
        )
    return x


def string(s: Any, data: dict[str, Any]) -> str:
    if not isinstance(s, str):
        return str(s)
    return re.sub(
        r"\{([^{}]+)\}", lambda match: str(eval_expression(match.group(1), data)), s
    )  # Match content inside curly braces


def many_scalars(x: cfg.ManyScalars, data: dict[str, Any]) -> pgt.ManyScalars:
    if isinstance(x, str):
        e = eval_expression(x, data)
        if tg.is_many_scalars(e):
            return e
        raise ValueError(
            f"Expression '{x}' evaluated to {e} of type {type(e)}, expected a sequence of scalars."
        )
    return [scalar(v, data) for v in x]


def one_or_many_scalars(
    x: cfg.OneOrManyScalars, data: dict[str, Any]
) -> pgt.OneOrManyScalars:
    if isinstance(x, str):
        e = eval_expression(x, data)
        if tg.is_one_or_many_scalars(e):
            return e
        raise ValueError(
            f"Expression '{x}' evaluated to {e} of type {type(e)}, expected a scalar or a sequence of them."
        )
    if isinstance(x, pgt.Scalar):
        return x
    return many_scalars(x, data)


def skyline_parameter(
    x: cfg.SkylineParameter, data: dict[str, Any]
) -> SkylineParameterLike:
    if isinstance(x, cfg.Scalar):
        return scalar(x, data)
    return SkylineParameter(
        value=many_scalars(x.value, data),
        change_times=many_scalars(x.change_times, data),
    )


def skyline_vector(
    x: cfg.SkylineVector, data: dict[str, Any]
) -> SkylineVectorCoercible:
    if isinstance(x, str):
        e = eval_expression(x, data)
        if tg.is_one_or_many_scalars(e):
            return e
        raise ValueError(
            f"Expression '{x}' evaluated to {e} of type {type(e)}, expected a SkylineVectorCoercible object (e.g., a scalar or a sequence of them)."
        )
    if isinstance(x, pgt.Scalar):
        return x
    if ctg.is_many_skyline_parameter_configs(x):
        return [skyline_parameter(p, data) for p in x]

    assert isinstance(x, cfg.SkylineVectorModel)

    change_times = many_scalars(x.change_times, data)
    if isinstance(x.value, str):
        e = eval_expression(x.value, data)
        if tg.is_many_one_or_many_scalars(e):
            value = e
        else:
            raise ValueError(
                f"Expression '{x.value}' evaluated to {e} of type {type(e)}, which cannot be coerced to a valid value for a SkylineVector (expected a sequence composed of scalars and/or sequences of scalars)."
            )
    else:
        value = [one_or_many_scalars(v, data) for v in x.value]

    if tg.is_many_scalars(value):
        return SkylineParameter(value=value, change_times=change_times)

    Ns = {len(elem) for elem in value if tg.is_many(elem)}
    if len(Ns) > 1:
        raise ValueError(
            f"All elements in the value of a SkylineVector config must be scalars or have the same length (config {x.value} yielded value={value} with inconsistent lengths {Ns})."
        )
    (N,) = Ns
    value = [[p] * N if isinstance(p, pgt.Scalar) else p for p in value]

    return SkylineVector(value=value, change_times=change_times)


def one_or_many_2D_scalars(
    x: cfg.OneOrMany2DScalars, data: dict[str, Any]
) -> pgt.OneOrMany2DScalars:
    if isinstance(x, str):
        e = eval_expression(x, data)
        if tg.is_one_or_many_2D_scalars(e):
            return e
        raise ValueError(
            f"Expression '{x}' evaluated to {e} of type {type(e)}, expected a nested (2D) sequence of scalars."
        )
    if isinstance(x, pgt.Scalar):
        return x
    return [many_scalars(v, data) for v in x]


def skyline_matrix(
    x: cfg.SkylineMatrix, data: dict[str, Any]
) -> SkylineMatrixCoercible | None:
    if x is None:
        return None

    if isinstance(x, str):
        e = eval_expression(x, data)
        if tg.is_one_or_many_2D_scalars(e):
            return e
        raise ValueError(
            f"Expression '{x}' evaluated to {e} of type {type(e)}, expected a SkylineMatrixCoercible object (e.g., a scalar or a nested (2D) sequence of them)."
        )
    if isinstance(x, pgt.Scalar):
        return x
    if ctg.is_many_skyline_vector_configs(x):
        return [skyline_vector(v, data) for v in x]

    assert isinstance(x, cfg.SkylineMatrixModel)

    change_times = many_scalars(x.change_times, data)
    if isinstance(x.value, str):
        e = eval_expression(x.value, data)
        if tg.is_many_one_or_many_2D_scalars(e):
            value = e
        else:
            raise ValueError(
                f"Expression '{x.value}' evaluated to {e} of type {type(e)}, which cannot be coerced to a valid value for a SkylineMatrix (expected a sequence composed of scalars and/or nested (2D) sequences of scalars)."
            )
    else:
        value = [one_or_many_2D_scalars(v, data) for v in x.value]

    if tg.is_many_scalars(value):
        return SkylineParameter(value=value, change_times=change_times)

    shapes: set[tuple[int, int]] = set()
    for elem in value:
        if tg.is_many_2D_scalars(elem):
            Ms = len(elem)
            Ns = {len(row) for row in elem}
            if len(Ns) > 1:
                raise ValueError(
                    f"The values of a SkylineMatrix config must be scalars or nested (2D) lists of them with a consistent row length (config {x.value} yielded element {elem} with row lengths {Ns})."
                )
            shapes.add((Ms, Ns.pop()))

    if len(shapes) > 1:
        raise ValueError(
            f"All elements in the value of a SkylineMatrix config must be scalars or nested (2D) lists of them with the same shape (config {x.value} yielded value={value} with inconsistent shapes {shapes})."
        )
    ((M, N),) = shapes
    value = [[[e] * N] * M if isinstance(e, pgt.Scalar) else e for e in value]

    return SkylineMatrix(value=value, change_times=change_times)


def distribution(x: cfg.Distribution, data: dict[str, Any]) -> cfg.Distribution:
    args = x.args
    for arg_name, arg_value in args.items():
        if isinstance(arg_value, str):
            args[arg_name] = eval_expression(arg_value, data)
    return cfg.Distribution(type=x.type, **args)


def data(context: dict[str, cfg.Distribution] | None, rng: Generator) -> dict[str, Any]:
    if context is None:
        return {}
    data: dict[str, Any] = {}
    for k, v in context.items():
        data[k] = np.array(distribution(v, data)(rng)).tolist()
    return data


def add_unbounded_population_timed_event(
    model: UnboundedPopulationModel,
    timed_event: cfg.UnboundedPopulationTimedEvent,
    data: dict[str, Any],
) -> None:
    model.add_timed_sampling_event(
        state=timed_event.state.format(**data),
        proportion=scalar(timed_event.proportion, data),
        removal=timed_event.removal,
        times=many_scalars(timed_event.times, data),
    )



from collections.abc import Callable, Iterator
from typing import Any, TypeGuard, Union, overload

import phylogenie._typeguards as tg
import phylogenie._typings as pgt
from phylogenie.skyline.parameter import SkylineParameter, is_skyline_parameter_like
from phylogenie.skyline.vector import (
    SkylineVector,
    SkylineVectorCoercible,
    SkylineVectorLike,
    SkylineVectorOperand,
    is_many_skyline_vectors_coercible,
    is_many_skyline_vectors_like,
    is_skyline_vector_like,
    is_skyline_vector_operand,
    skyline_vector,
)

SkylineMatrixOperand = Union[SkylineVectorOperand, "SkylineMatrix"]
SkylineMatrixCoercible = Union[pgt.OneOrMany[SkylineVectorCoercible], "SkylineMatrix"]


def is_skyline_matrix_operand(x: Any) -> TypeGuard[SkylineMatrixOperand]:
    return isinstance(x, SkylineMatrix) or is_skyline_vector_operand(x)


class SkylineMatrix:
    """
    Represent a matrix of skyline vectors.
    """

    def __init__(
        self,
        params: pgt.Many[SkylineVectorLike] | None = None,
        value: pgt.Many3DScalars | None = None,
        change_times: pgt.ManyScalars | None = None,
    ):
        """
        Initialize a SkylineMatrix from a sequence of skyline vectors or from value and change times.

        Parameters
        -----------
        params : Many[SkylineVectorLike] | None, optional
            Sequence of skyline vectors or sequences to compose the matrix.
        value : Many3DScalars | None, optional
            3D sequence of values for each parameter over time.
        change_times : ManyScalars | None, optional
            Times at which the parameter values change. Must be provided if `value` is provided.
        """
        if params is not None and value is None and change_times is None:
            if is_many_skyline_vectors_like(params):
                self._params = [
                    p if isinstance(p, SkylineVector) else SkylineVector(p)
                    for p in params
                ]
            else:
                raise TypeError(
                    f"It is impossible to create a SkylineMatrix from `params` {params} of type {type(params)}. Please provide a sequence composed of SkylineVectorLike objects (a SkylineVectorLike object can either be a SkylineVector or a sequence of scalars and/or SkylineParameters)."
                )
            lengths = {len(p) for p in self._params}
            if len(lengths) > 1:
                raise ValueError(
                    f"All `params` must have the same length to create a SkylineMatrix (got params={params} with lengths {lengths})."
                )
        elif params is None and value is not None and change_times is not None:
            if tg.is_many_3D_scalars(value):
                lengths = {len(matrix) for matrix in value}
                if len(lengths) > 1:
                    raise ValueError(
                        f"All matrices in the `value` of a SkylineMatrix must have the same number of rows (got matrices={value} with row lengths {lengths})."
                    )
            else:
                raise TypeError(
                    f"It is impossible to create a SkylineMatrix from `value` {value} of type {type(value)}. Please provide a nested (3D) sequence of scalar values."
                )
            self._params = [
                SkylineVector(
                    value=[matrix[i] for matrix in value], change_times=change_times
                )
                for i in range(len(value[0]))
            ]
        else:
            raise ValueError(
                "Either `params` or both `value` and `change_times` must be provided to create a SkylineMatrix."
            )

    @property
    def params(self) -> tuple[SkylineVector, ...]:
        """
        Return the skyline vectors composing this matrix.

        Returns
        --------
        tuple[SkylineVector, ...]
            Skyline vectors for each row.
        """
        return tuple(self._params)

    @property
    def n_rows(self) -> int:
        """
        Return the number of rows in the matrix.

        Returns
        --------
        int
            Number of rows.
        """
        return len(self.params)

    @property
    def n_cols(self) -> int:
        """
        Return the number of columns in the matrix.

        Returns
        --------
        int
            Number of columns.
        """
        return len(self.params[0])

    @property
    def shape(self) -> tuple[int, int]:
        """
        Return the matrix shape as (n_rows, n_cols).

        Returns
        --------
        tuple[int, int]
            Matrix shape.
        """
        return self.n_rows, self.n_cols

    @property
    def change_times(self) -> pgt.Vector1D:
        """
        Return the union of change times across all rows.

        Returns
        --------
        Vector1D
            Sorted unique change times.
        """
        return tuple(sorted(set([t for row in self.params for t in row.change_times])))

    @property
    def value(self) -> pgt.Vector3D:
        """
        Return the value tensor over time.

        Returns
        --------
        Vector3D
            Tensor where each slice corresponds to a time segment.
        """
        return tuple(self.get_value_at_time(t) for t in (0, *self.change_times))

    def get_value_at_time(self, time: pgt.Scalar) -> pgt.Vector2D:
        """
        Evaluate the matrix at a given time.

        Parameters
        -----------
        time : Scalar
            Time at which to evaluate.

        Returns
        --------
        Vector2D
            Matrix values at the given time.
        """
        return tuple(param.get_value_at_time(time) for param in self.params)

    def _operate(
        self,
        other: SkylineMatrixOperand,
        func: Callable[
            [SkylineVector, SkylineVector | SkylineParameter], SkylineVector
        ],
    ) -> "SkylineMatrix":
        if is_skyline_matrix_operand(other):
            other = skyline_matrix(other, self.n_rows, self.n_cols)
        elif isinstance(other, SkylineMatrix):
            if other.shape != self.shape:
                raise ValueError(
                    f"It is impossible to operate on SkylineMatrices of different shapes (got self={self.shape} and other={other.shape})."
                )
        else:
            return NotImplemented
        return SkylineMatrix(
            [func(p1, p2) for p1, p2 in zip(self.params, other.params)]
        )

    def __add__(self, operand: SkylineMatrixOperand) -> "SkylineMatrix":
        return self._operate(operand, lambda x, y: x + y)

    def __radd__(self, operand: SkylineVectorOperand) -> "SkylineMatrix":
        return self._operate(operand, lambda x, y: y + x)

    def __sub__(self, operand: SkylineMatrixOperand) -> "SkylineMatrix":
        return self._operate(operand, lambda x, y: x - y)

    def __rsub__(self, operand: SkylineVectorOperand) -> "SkylineMatrix":
        return self._operate(operand, lambda x, y: y - x)

    def __mul__(self, operand: SkylineMatrixOperand) -> "SkylineMatrix":
        return self._operate(operand, lambda x, y: x * y)

    def __rmul__(self, operand: SkylineVectorOperand) -> "SkylineMatrix":
        return self._operate(operand, lambda x, y: y * x)

    def __truediv__(self, operand: SkylineMatrixOperand) -> "SkylineMatrix":
        return self._operate(operand, lambda x, y: x / y)

    def __rtruediv__(self, operand: SkylineVectorOperand) -> "SkylineMatrix":
        return self._operate(operand, lambda x, y: y / x)

    @property
    def T(self) -> "SkylineMatrix":
        """
        Return the transpose of the skyline matrix.

        Returns
        --------
        SkylineMatrix
            Transposed matrix.
        """
        return SkylineMatrix([[v[i] for v in self] for i in range(self.n_cols)])

    def __bool__(self) -> bool:
        return any(self.params)

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, SkylineMatrix) and self.params == other.params

    def __repr__(self) -> str:
        return f"SkylineMatrix(value={list(self.value)}, change_times={list(self.change_times)})"

    def __iter__(self) -> Iterator[SkylineVector]:
        return iter(self.params)

    def __len__(self) -> int:
        return self.n_rows

    @overload
    def __getitem__(self, item: int) -> SkylineVector: ...
    @overload
    def __getitem__(self, item: slice) -> "SkylineMatrix": ...
    @overload
    def __getitem__(self, item: tuple[int, int]) -> SkylineParameter: ...
    @overload
    def __getitem__(self, item: tuple[int, slice]) -> SkylineVector: ...
    @overload
    def __getitem__(self, item: tuple[slice, int]) -> SkylineVector: ...
    @overload
    def __getitem__(self, item: tuple[slice, slice]) -> "SkylineMatrix": ...
    def __getitem__(
        self, item: int | slice | tuple[int | slice, int | slice]
    ) -> Union[SkylineParameter | SkylineVector, "SkylineMatrix"]:
        if isinstance(item, int):
            return self.params[item]
        if isinstance(item, slice):
            return SkylineMatrix(self.params[item])
        row_idx, col_idx = item
        if isinstance(row_idx, int):
            return self.params[row_idx][col_idx]
        if isinstance(col_idx, int):
            return SkylineVector([row[col_idx] for row in self.params[row_idx]])
        return SkylineMatrix([row[col_idx] for row in self.params[row_idx]])

    def __setitem__(self, item: int, value: SkylineVectorLike) -> None:
        if not is_skyline_vector_like(value):
            raise TypeError(
                f"It is impossible to set item of SkylineMatrix to value {value} of type {type(value)}. Please provide a SkylineVectorLike object (i.e., a SkylineVector or a sequence of scalars and/or SkylineParameters)."
            )
        self._params[item] = skyline_vector(value, self.n_cols)


def skyline_matrix(
    x: SkylineMatrixCoercible, n_rows: int, n_cols: int
) -> SkylineMatrix:
    """
    Coerce a value into a SkylineMatrix of a specific shape.

    Parameters
    -----------
    x : SkylineMatrixCoercible
        An object to coerce into a SkylineMatrix.
    n_rows : int
        Expected number of rows.
    n_cols : int
        Expected number of columns.

    Returns
    --------
    SkylineMatrix
        Coerced skyline matrix with the requested shape.
    """
    if n_rows <= 0 or n_cols <= 0:
        raise ValueError(
            f" n_rows and n_cols must be positive integers to create a SkylineMatrix (got n_rows={n_rows} and n_cols={n_cols})."
        )

    if is_skyline_parameter_like(x):
        return SkylineMatrix([[x] * n_cols] * n_rows)
    if is_skyline_vector_like(x) or is_many_skyline_vectors_coercible(x):
        if len(x) == n_rows:
            return SkylineMatrix([skyline_vector(p, n_cols) for p in x])
        elif len(x) == n_cols:
            return SkylineMatrix([skyline_vector(p, n_rows) for p in x]).T
        raise ValueError(
            f"Expected a SkylineVectorLike of size {n_rows} or {n_cols}, got {x} of size {len(x)}."
        )

    if not isinstance(x, SkylineMatrix):
        raise TypeError(
            f"It is impossible to coerce {x} of type {type(x)} into a SkylineMatrix. Please provide either:\n"
            "- a SkylineMatrix,\n"
            "- a SkylineVectorCoercible object (i.e., a scalar, a SkylineParameter, a SkylineVector, or a sequence of scalars and/or SkylineParameters),\n"
            "- a sequence of SkylineVectorCoercible objects."
        )

    if x.shape != (n_rows, n_cols):
        raise ValueError(
            f"Expected an SkylineMatrix of shape ({n_rows}, {n_cols}), got {x} of shape {x.shape}."
        )

    return x



import re
from abc import ABC, abstractmethod
from collections import defaultdict
from math import comb

import numpy as np
from numpy.random import default_rng

from phylogenie._mixins import MetadataMixin
from phylogenie.tree_node import TreeNode
from phylogenie.treesimulator.events import StochasticEvent, TimedEvent


class Model(MetadataMixin, ABC):
    def __init__(self, max_time: float | None = None, seed: int | None = None):
        super().__init__()
        self.max_time = max_time
        self._rng = default_rng(seed)
        self._stochastic_events: list[StochasticEvent] = []
        self._timed_events: list[TimedEvent] = []

    def init(self):
        self.clear()
        self._current_time = 0.0
        self._next_individual_id = 0
        self._population: dict[str, set[int]] = defaultdict(set)
        self._states: dict[int, str] = {}

    def create_new_individual(self, state: str) -> int:
        self._next_individual_id += 1
        self._population[state].add(self._next_individual_id)
        self._states[self._next_individual_id] = state
        return self._next_individual_id

    def remove_individual(self, individual: int) -> None:
        state = self._states[individual]
        self._population[state].remove(individual)
        del self._states[individual]

    def migrate_individual(self, individual: int, new_state: str) -> None:
        old_state = self._states[individual]
        self._population[old_state].remove(individual)
        self._states[individual] = new_state
        self._population[new_state].add(individual)

    def get_individuals(self, state: str) -> list[int]:
        return [
            individual
            for s, individuals in self._population.items()
            if re.fullmatch(state, s) is not None
            for individual in individuals
        ]

    def count_individuals(self, state: str) -> int:
        return len(self.get_individuals(state))

    def draw_individual(self, state: str) -> int:
        return self._rng.choice(self.get_individuals(state))

    def draw_individuals(self, count: int, state: str) -> list[int]:
        return self._rng.choice(
            self.get_individuals(state), size=count, replace=False
        ).tolist()

    @property
    @abstractmethod
    def tree_size(self) -> int: ...

    @abstractmethod
    def get_tree(self) -> TreeNode | None: ...

    # -------------------------
    # Gillespie-related methods
    # -------------------------

    def _get_propensity(self, event: StochasticEvent, time: float) -> float:
        propensity = event.rate.get_value_at_time(time)
        for state, count in event.reactants.items():
            propensity *= comb(self.count_individuals(state), count)
        return propensity

    def _get_next_change_time(self) -> float | None:
        change_times = [
            t
            for e in self._stochastic_events
            for t in e.rate.change_times
            if t > self._current_time
        ]
        return min(change_times) if change_times else None

    def _get_next_stochastic_event(
        self,
    ) -> tuple[float, StochasticEvent] | tuple[None, None]:
        propensities = [
            self._get_propensity(e, self._current_time) for e in self._stochastic_events
        ]
        total_propensity = sum(propensities)
        if not total_propensity:
            return None, None
        next_time = self._current_time + self._rng.exponential(1 / total_propensity)
        event_idx = np.searchsorted(
            np.cumsum(propensities) / total_propensity, self._rng.random()
        )
        return next_time, self._stochastic_events[int(event_idx)]

    def _get_next_timed_events(
        self,
    ) -> tuple[float, list[TimedEvent]] | tuple[None, None]:
        times = [
            t for e in self._timed_events for t in e.times if t > self._current_time
        ]
        if not times:
            return None, None
        next_time = min(times)
        return next_time, [e for e in self._timed_events if next_time in e.times]

    def step(self) -> bool:
        stochastic_event_time, stochastic_event = self._get_next_stochastic_event()
        timed_events_time, timed_events = self._get_next_timed_events()
        next_change_time = self._get_next_change_time()

        # If there are no more events to process, we stop the simulation.
        if all(
            t is None
            for t in [next_change_time, timed_events_time, stochastic_event_time]
        ):
            return False

        # If the next event is beyond the max time, we stop the simulation.
        if self.max_time is not None and all(
            t is None or self.max_time < t
            for t in [
                stochastic_event_time,
                timed_events_time,
                next_change_time,
            ]
        ):
            return False

        # If the next change time is the earliest, we advance time to it.
        if next_change_time is not None and all(
            t is None or next_change_time <= t
            for t in [timed_events_time, stochastic_event_time]
        ):
            self._current_time = next_change_time

        # If the next timed events time is the earliest, we process them.
        if timed_events_time is not None and all(
            t is None or timed_events_time <= t
            for t in [next_change_time, stochastic_event_time]
        ):
            self._current_time = timed_events_time
            for event in timed_events:  # pyright: ignore
                event.fn(self._current_time)

        # If the next stochastic event time is the earliest, we process it.
        if stochastic_event_time is not None and all(
            t is None or stochastic_event_time <= t
            for t in [next_change_time, timed_events_time]
        ):
            self._current_time = stochastic_event_time
            stochastic_event.fn(self._current_time)  # pyright: ignore

        return True



from collections.abc import Iterable, Iterator
from dataclasses import dataclass
from datetime import date

import numpy as np


@dataclass
class Sequence:
    """
    Represent a single aligned sequence with optional sampling time.

    Parameters
    -----------
    id : str
        Identifier for the sequence.
    chars : str
        Aligned sequence characters.
    time : float | date | None, optional
        Sampling time associated with the sequence.
    """

    id: str
    chars: str
    time: float | date | None = None

    def __len__(self) -> int:
        return len(self.chars)


class MSA:
    """
    Store and validate a multiple sequence alignment.

    The alignment ensures that all sequences have identical length and provides
    convenient accessors for IDs, times, and alignment shape.
    """

    def __init__(self, sequences: Iterable[Sequence]):
        """
        Initialize the MSA with a collection of sequences.

        Parameters
        -----------
        sequences : Iterable[Sequence]
            Sequence records to include in the alignment.
        """
        self._sequences = list(sequences)
        lengths = {len(sequence) for sequence in sequences}
        if len(lengths) > 1:
            raise ValueError(
                f"All sequences in the alignment must have the same length (got lengths: {lengths})"
            )

    @property
    def sequences(self) -> tuple[Sequence, ...]:
        """
        Return the alignment sequences as a tuple.

        Returns
        --------
        tuple[Sequence, ...]
            The sequences in the alignment.
        """
        return tuple(self._sequences)

    @property
    def ids(self) -> list[str]:
        """
        Return the sequence identifiers.

        Returns
        --------
        list[str]
            Sequence IDs in the order stored.
        """
        return [sequence.id for sequence in self.sequences]

    @property
    def times(self) -> list[float | date]:
        """
        Return sampling times for all sequences.

        Raises a ValueError if any sequence is missing a sampling time.

        Returns
        --------
        list[float | date]
            Sampling times for each sequence.
        """
        times: list[float | date] = []
        for sequence in self:
            if sequence.time is None:
                raise ValueError(f"Time is not set for sequence {sequence.id}.")
            times.append(sequence.time)
        return times

    @property
    def alignment(self) -> list[list[str]]:
        """
        Return the alignment as a list of character lists.

        Returns
        --------
        list[list[str]]
            The alignment matrix where rows are sequences.
        """
        return [list(sequence.chars) for sequence in self.sequences]

    @property
    def n_sequences(self) -> int:
        """
        Return the number of sequences in the alignment.

        Returns
        --------
        int
            Number of sequences.
        """
        return len(self.sequences)

    @property
    def n_sites(self) -> int:
        """
        Return the number of aligned sites per sequence.

        Returns
        --------
        int
            Sequence length in sites.
        """
        return len(self.alignment[0])

    @property
    def shape(self) -> tuple[int, int]:
        """
        Return the alignment shape as (n_sequences, n_sites).

        Returns
        --------
        tuple[int, int]
            Alignment shape.
        """
        return self.n_sequences, self.n_sites

    def count_informative_sites(self) -> int:
        """
        Count phylogenetically informative sites in the alignment.

        A site is informative if it has at least two character states each
        appearing in at least two sequences.

        Returns
        --------
        int
            Number of informative sites.
        """
        n_informative_sites = 0
        for column in np.array(self.alignment).T:
            column: np.typing.NDArray[np.str_]
            _, char_counts = np.unique(column, return_counts=True)
            is_informative_char = char_counts >= 2
            if (is_informative_char).sum() >= 2:
                n_informative_sites += 1
        return n_informative_sites

    def count_unique_sequences(self) -> int:
        """
        Count the number of unique sequences in the alignment.

        Returns
        --------
        int
            Count of distinct sequences.
        """
        return len(np.unique(self.alignment, axis=0))

    def __len__(self) -> int:
        """
        Return the number of sequences in the alignment.

        Returns
        --------
        int
            Number of sequences in the alignment.
        """
        return self.n_sequences

    def __getitem__(self, item: int) -> Sequence:
        """
        Retrieve a sequence by index.

        Parameters
        -----------
        item : int
            Index of the sequence to retrieve.

        Returns
        --------
        Sequence
            The sequence at the specified index.
        """
        return self.sequences[item]

    def __iter__(self) -> Iterator[Sequence]:
        """
        Iterate over the sequences in the alignment.

        Returns
        --------
        Iterator[Sequence]
            An iterator over the alignment sequences.
        """
        return iter(self.sequences)



import os
from abc import ABC, abstractmethod
from enum import Enum
from typing import Any

import joblib
import pandas as pd
from numpy.random import Generator, default_rng
from tqdm import tqdm

from phylogenie.generators._configs import Distribution, StrictBaseModel


class DataType(str, Enum):
    TREES = "trees"
    MSAS = "msas"


DATA_DIRNAME = "data"
METADATA_FILENAME = "metadata.csv"


class DatasetGenerator(ABC, StrictBaseModel):
    output_dir: str = "phylogenie-outputs"
    n_samples: int | dict[str, int] = 1
    n_jobs: int = -1
    seed: int | None = None
    context: dict[str, Distribution] | None = None

    @abstractmethod
    def generate_one(
        self,
        filename: str,
        context: dict[str, Distribution] | None = None,
        seed: int | None = None,
    ) -> dict[str, Any]: ...

    def _generate(self, rng: Generator, n_samples: int, output_dir: str) -> None:
        if os.path.exists(output_dir):
            print(f"Output directory {output_dir} already exists. Skipping.")
            return

        data_dir = (
            output_dir
            if self.context is None
            else os.path.join(output_dir, DATA_DIRNAME)
        )
        os.makedirs(data_dir)

        jobs = joblib.Parallel(n_jobs=self.n_jobs, return_as="generator_unordered")(
            joblib.delayed(self.generate_one)(
                seed=int(rng.integers(2**32)),
                filename=os.path.join(data_dir, str(i)),
                context=self.context,
            )
            for i in range(n_samples)
        )
        df = pd.DataFrame(
            [j for j in tqdm(jobs, f"Generating {data_dir}...", n_samples)]
        )
        df.to_csv(os.path.join(output_dir, METADATA_FILENAME), index=False)

    def generate(self) -> None:
        rng = default_rng(self.seed)
        if isinstance(self.n_samples, dict):
            for key, n_samples in self.n_samples.items():
                output_dir = os.path.join(self.output_dir, key)
                self._generate(rng, n_samples, output_dir)
        else:
            self._generate(rng, self.n_samples, self.output_dir)
